from functools import partial

problems = {}

# a decorator to store functions in problems as we go
def add_to_problems(shortname, function):
    """
    add a problem to the problems dict
    """
    name = function.__name__.replace("_", "-")
    def wrapped():
        raw = function()
        raw['shortname'] = shortname
        return raw
    problems[name] = wrapped
    return wrapped


@partial(add_to_problems, "short")
def knuth_original():
    return dict(value="Bingo")

print(problems['knuth-original']())



"""
see README.md for more details
"""

import time
from pathlib import Path
from argparse import ArgumentParser

import pandas as pd


# samples
import exact_cover_samples as ecs

# exact-cover-py
from exact_cover_py import exact_covers, __version__ as exact_cover_py_version

# exact cover
# from exact_cover import get_exact_cover, __version__ as exact_cover_version
from exact_cover import get_exact_cover, get_all_solutions as get_exact_covers
exact_cover_version = "1.4.0a1"
print(f"WARNING using hard-wired version {exact_cover_version=}")

# xcover
from xcover import covers_bool, __version__ as xcover_version


# accumulate results in this file
RESULTS = "results.csv"


def store_records(records):
    if Path(RESULTS).exists():
        past_records = pd.read_csv(RESULTS)
    else:
        past_records = pd.DataFrame()
    new_results = pd.DataFrame(records)
    pd.concat([past_records, new_results], ignore_index=True).to_csv(RESULTS, index=False)


# run the algorithms, each lib has its own logic
class Library:
    def run(self, problem, size):
        raise NotImplementedError

class ExactCover(Library):
    def __init__(self):
        self.name = "exact-cover"
        self.version = exact_cover_version
    def run(self, problem, size):
        if size == 1:
            r = [get_exact_cover(problem["data"])]
        elif size == 0:
            r = get_exact_covers(problem["data"])
        else:
            r = get_exact_covers(problem["data"], max_count=size)
        return len(r)

class ExactCoverPy(Library):
    def __init__(self):
        self.name = "exact-cover-py"
        self.version = exact_cover_py_version
    def run(self, problem, size):
        solutions = exact_covers(problem["data"])
        if size > 0:
            counter = 0
            try:
                for _ in range(size):
                    next(solutions)
                    counter += 1
            except StopIteration:
                pass
        elif size == 0:
            counter = 0
            for _ in solutions:
                counter += 1
        return counter

class XCover(Library):
    def __init__(self):
        self.name = "xcover"
        self.version = xcover_version
    def run(self, problem, size):
        solutions = covers_bool(problem["data"])
        if size > 0:
            counter = 0
            try:
                for _ in range(size):
                    next(solutions)
                    counter += 1
            except StopIteration:
                pass
        elif size == 0:
            counter = 0
            for _ in solutions:
                counter += 1
        return counter

ALL_LIBS = [ExactCover(), XCover(), ExactCoverPy()]


def run_library(lib, problem, run_index, size):
    requested = size
    expected = len(problem["solutions"])
    if size != 0:
        expected = min(size, expected)
    start = time.perf_counter()
    try:
        computed = lib.run(problem, size)
        end = time.perf_counter()
        result = dict(
            library=lib.name, version=lib.version,
            problem=problem['shortname'], run=run_index,
            requested=requested, expected=expected, computed=computed,
            time=end - start, error="")
        if computed != expected:
            result["error"] = f"expected {expected} but got {computed}"
        return result
    except Exception as e:
        return dict(
            library=lib.name, version=lib.version,
            problem=problem['shortname'], run=run_index,
            requested=requested, expected=expected, computed=-1,
            time=0.0, error=str(e))



def run(algo, full, runs):
    # the various configurations we try
    # sizes means how many solutions we want to compute
    # with 0 meaning all of them
    # passing full=False will skip the 0 case
    sizes = [1, 50]
    if full:
        sizes.append(0)
    if algo == -1:
        libs = ALL_LIBS
    elif 0 <= algo < len(ALL_LIBS):
        libs = [ALL_LIBS[algo]]
    else:
        raise ValueError(f"invalid algo {algo}")
    problems = ecs.problems
    for problem_fn in problems.values():
        problem = problem_fn()
        print(f"=== problem {problem['name']}")
        for lib in libs:
            for run_index in range(runs):
                print(f"= library {lib.name} (run {run_index})")
                store_records(
                    [run_library(lib, problem, run_index, size)
                     for size in sizes])

def main():
    parser = ArgumentParser()
    parser.add_argument("-r", "--runs", type=int, default=1)
    parser.add_argument("-f", "--full", action="store_true", default=False)
    parser.add_argument("-a", "--algo", action="store",
                        type=int, default=-1,
                        help="0: exact-cover, 1: xcover, 2: exact-cover-py")
    args = parser.parse_args()
    run(args.algo, args.full, args.runs)

if __name__ == "__main__":
    main()



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: all,-hidden,-heading_collapsed,-run_control,-trusted
#     formats: py:percent
#     notebook_metadata_filter: all, -jupytext.text_representation.jupytext_version,
#       -jupytext.text_representation.format_version,-language_info.version, -language_info.codemirror_mode.version,
#       -language_info.codemirror_mode,-language_info.file_extension, -language_info.mimetype,
#       -toc, -rise, -version
#     text_representation:
#       extension: .py
#       format_name: percent
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
#   language_info:
#     name: python
#     nbconvert_exporter: python
#     pygments_lexer: ipython3
# ---

# %%
import pandas as pd
import matplotlib.pyplot as plt

# %%
# %matplotlib ipympl

# %% [markdown]
# # benchmarking

# %% [markdown]
# ## loading results

# %%
df = pd.read_csv("results.csv")
df.shape

# %%
df.head()

# %% [markdown]
# ## errors

# %%
df[df.computed == -1]

# %%
df[df.error.notna()]

# %% [markdown]
# ## keeping only run 1

# %%
df = df[df.run == 1]
df.shape

# %% [markdown]
# ## artificially divide exact-cover-py by 50

# %%
time_series = df.time
updated = time_series[df.library == "exact-cover-py"] / 50
time_series.update(updated)

library_series = df.library
updated = library_series[df.library == "exact-cover-py"]
updated[:] = "exact-cover-py / 50"
library_series.update(updated)

# %% [markdown]
# ## drawing

# %%
keep = "library+problem+requested+time".split("+")

table = df[keep].pivot_table(columns=["requested", "library"], index="problem", values="time")
table

# %%
df_all_solutions = table.loc[:, 0]
df_all_solutions.plot()
plt.savefig("results.svg", format="svg")
plt.savefig("results.png", format="png")



"""
this module mainly exposes a dictionary 'problems'
of the form problem_name -> problem_function

each problem function returns a dict with 2 keys
they should return a dict with
- data: a numpy array
- solutions: purpusefully left loosely specified as a collection of solutions
  e.g. may be a list of lists, a numpy array, etc..
  tentatively we return them in the same order as the
  original article with the S heuristic

also there a are 3 helper tools that help canocalize solutions for comparison
"""

from functools import partial
from importlib import resources

import numpy as np
import pandas as pd

DTYPE_FOR_ARRAY = bool


# canonicalize solutions
def canonical_1(solution):
    """
    how to canonicalize one solution
    """
    return tuple(sorted(solution))

def canonical_s(solutions):
    """
    apply canonical_1 on all solutions, as a list in the original order
    """
    return [canonical_1(solution) for solution in solutions]

def canonical(solutions):
    """
    same but also turn into a set
    """
    return set(canonical_s(solutions))


# locate and load packaged files
def load_npy(filename):
    p = resources.files("exact_cover_samples.data").joinpath(f"{filename}.npy")
    return np.load(str(p))

def load_csv(filename):
    p = resources.files("exact_cover_samples.data").joinpath(f"{filename}.csv")
    return pd.read_csv(p, header=None).to_numpy()


problems = {}

# a decorator to store functions in problems as we go
# a decorator to store functions in problems as we go
def add_to_problems(shortname, function):
    """
    add a problem to the problems dict
    """
    name = function.__name__.replace("_", "-")
    def wrapped():
        raw = function()
        raw['name'] = name
        raw['shortname'] = shortname
        return raw
    problems[shortname] = wrapped
    return wrapped


# may be useful to test the algorithm on a trivial problem
# since this is the one illustrated in the original article
@partial(add_to_problems, "knuth")
def knuth_original():
    to_cover = np.array(
        [
            [0, 0, 1, 0, 1, 1, 0],
            [1, 0, 0, 1, 0, 0, 1],
            [0, 1, 1, 0, 0, 1, 0],
            [1, 0, 0, 1, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 1],
            [0, 0, 0, 1, 1, 0, 1],
        ]
    )
    return {"data": to_cover, "solutions": [(0, 3, 4)]}


# same problem in fact, but expressed a little differently
# https://en.wikipedia.org/wiki/Exact_cover#Detailed_example
@partial(add_to_problems, "wikip")
def detailed_wikipedia():
    sets = [
        {1, 4, 7},
        {1, 4},  # <- 1
        {4, 5, 7},
        {3, 5, 6},  # <- 3
        {2, 3, 6, 7},
        {2, 7},  # <- 5
    ]
    return dict(
        data=np.array(
            [[1 if i in s else 0 for i in range(1, 8)] for s in sets],
            dtype=DTYPE_FOR_ARRAY,
        ),
        solutions=[(1, 3, 5)],
    )


@partial(add_to_problems, "brtf1")
def bruteforce1():
    to_cover = [
        [1, 0, 0, 1, 0, 0, 1, 0],  # <- sol1
        [0, 1, 0, 0, 1, 0, 0, 1],  # <- sol1
        [0, 0, 1, 0, 0, 1, 0, 0],  # <- sol1
        [0, 0, 0, 1, 0, 0, 0, 0],  # <- sol2
        [1, 0, 1, 0, 1, 0, 0, 1],  # <- sol2
        [0, 1, 0, 0, 0, 1, 1, 0],  # <- sol2
    ]
    return dict(
        data=np.array(to_cover, dtype=DTYPE_FOR_ARRAY), solutions=[(0, 1, 2), (3, 4, 5)]
    )


@partial(add_to_problems, "brtf2")
def bruteforce2():
    to_cover = [
        [1, 0, 0, 1, 0, 0, 1, 0],  # <- sol1
        [0, 1, 0, 0, 1, 0, 0, 1],  # <- sol1
        [0, 0, 1, 0, 0, 1, 0, 0],  # <- sol1
        [0, 0, 0, 1, 0, 0, 0, 0],  # <- sol2
        [1, 0, 1, 0, 1, 0, 0, 1],  # <- sol2
        [0, 1, 0, 0, 0, 1, 1, 0],  # <- sol2
        [1, 0, 0, 1, 0, 0, 1, 0],  # <- sol1
        [0, 1, 0, 0, 1, 0, 0, 1],  # <- sol1
        [0, 0, 1, 0, 0, 1, 0, 0],  # <- sol1
    ]
    return dict(
        data=np.array(to_cover, dtype=DTYPE_FOR_ARRAY),
        solutions=canonical(
            [
                (0, 1, 2),
                (0, 1, 8),
                (0, 7, 2),
                (0, 7, 8),
                (4, 5, 3),
                (6, 1, 2),
                (6, 1, 8),
                (6, 7, 2),
                (6, 7, 8),
            ]
        ),
    )


@partial(add_to_problems, "btrf3")
def bruteforce3():
    to_cover = [
        [1, 0, 0, 1, 0, 0, 1, 0],  # <- sol1
        [0, 1, 0, 0, 1, 0, 0, 1],  # <- sol1
        [0, 0, 1, 0, 0, 1, 0, 0],  # <- sol1
        [0, 0, 0, 1, 0, 0, 0, 0],  # <- sol2
        [1, 0, 1, 0, 1, 0, 0, 1],  # <- sol2
        [0, 1, 0, 0, 0, 1, 1, 0],  # <- sol2
        [1, 0, 0, 1, 0, 0, 1, 0],  # <- sol1
        [0, 1, 0, 0, 1, 0, 0, 1],  # <- sol1
        [0, 0, 1, 0, 0, 1, 0, 0],  # <- sol1
        [0, 0, 0, 1, 0, 0, 0, 0],  # <- sol2
        [1, 0, 1, 0, 1, 0, 0, 1],  # <- sol2
        [0, 1, 0, 0, 0, 1, 1, 0],  # <- sol2
    ]
    return dict(
        data=np.array(to_cover, dtype=DTYPE_FOR_ARRAY),
        solutions=canonical(
            [
                (0, 1, 2),
                (0, 1, 8),
                (0, 7, 2),
                (0, 7, 8),
                (4, 5, 3),
                (4, 5, 9),
                (4, 11, 3),
                (4, 11, 9),
                (6, 1, 2),
                (6, 1, 8),
                (6, 7, 2),
                (6, 7, 8),
                (10, 5, 3),
                (10, 5, 9),
                (10, 11, 3),
                (10, 11, 9),
            ]
        ),
    )


@partial(add_to_problems, "btrf30")
def bruteforce3_odd_zeros():
    p = bruteforce3()
    d, s = p["data"], p["solutions"]
    r, c = d.shape
    # add same area of 0s on the right hand side of d
    d1 = np.hstack((d, np.zeros(d.shape, dtype=d.dtype)))
    # reshape it - each gets folded in 2
    # so we end up with all the odd rows being 0
    d2 = d1.reshape((-1, c))
    # non empty rows are now 0 2 4, so twice the original index
    s = {tuple(map(lambda i: i * 2, t)) for t in s}
    return dict(data=d2, solutions=s)


@partial(add_to_problems, "btrf31")
def bruteforce3_even_zeros():
    p = bruteforce3()
    d, s = p["data"], p["solutions"]
    r, c = d.shape
    # add same area of 0s on the left hand side of d
    d1 = np.hstack((np.zeros(d.shape, dtype=d.dtype), d))
    # reshape it - each gets folded in 2
    # so we end up with all the even rows being 0
    d2 = d1.reshape((-1, c))
    # non empty rows are now 1 3 5, so twice the original index + 1
    s = {tuple(map(lambda i: i * 2 + 1, t)) for t in s}
    return dict(data=d2, solutions=s)


# problem originally based on solving the trivial problem
# of arranging 2 identical triminos on a 3x3 board

#    +--+
#    |  |
# +--+--+
# |  |  |
# +--+--+

# +--+--+--+
# |xx|  |xx|
# +--+--+--+
# |  |  |  |
# +--+--+--+
# |xx|  |  |
# +--+--+--+


# this problem has 2 solutions
# (5, 13) and (6, 12)
@partial(add_to_problems, "strim")
def small_trimino():
    to_cover = [
        [1, 0, 0, 1, 1, 0, 1, 0],
        [1, 0, 0, 0, 1, 1, 0, 1],
        [1, 0, 0, 0, 1, 1, 1, 0],
        [1, 0, 1, 0, 1, 1, 0, 0],
        [1, 0, 0, 0, 1, 0, 1, 1],
        [1, 0, 1, 1, 1, 0, 0, 0],  # <- 5
        [1, 0, 0, 0, 0, 1, 1, 1],  # <- 6
        [0, 1, 0, 1, 1, 0, 1, 0],
        [0, 1, 0, 0, 1, 1, 0, 1],
        [0, 1, 0, 0, 1, 1, 1, 0],
        [0, 1, 1, 0, 1, 1, 0, 0],
        [0, 1, 0, 0, 1, 0, 1, 1],
        [0, 1, 1, 1, 1, 0, 0, 0],  # <- 12
        [0, 1, 0, 0, 0, 1, 1, 1],  # <- 13
    ]
    return dict(
        data=np.array(to_cover, dtype=DTYPE_FOR_ARRAY),
        solutions=[(5, 13), (6, 12)],
    )


@partial(add_to_problems, "strimf")
def small_trimino_from_file():
    return dict(
        data=load_npy("small-trimino"),
        solutions=[(5, 13), (6, 12)],
    )


@partial(add_to_problems, "pchess")
def pentomino_chessboard():
    to_cover = load_csv("pentominos-chessboard")
    solutions = load_csv("pentominos-chessboard-solutions")

    return dict(
        data=to_cover,
        solutions=solutions,
    )


@partial(add_to_problems, "p5x12")
def pentomino_5x12():
    to_cover = load_csv("pentominos-5-12")
    solutions = load_csv("pentominos-5-12-solutions")
    return dict(
        data=to_cover,
        solutions=solutions,
    )


def print_problem(problem, header=False):
    name = problem["name"]
    shortname = problem["shortname"]
    data = problem["data"]
    solutions = problem["solutions"]
    if header:
        print(f"{' '+shortname+' ':=^50}")
        repeat = ""
    else:
        repeat = f"short={shortname} "
    print(
            f"{repeat}"
            f"size = {data.shape}, "
            f" {len(canonical(solutions))} solutions"
            f" full_name={name}")


def summary(filter=None):
    """
    convenience to display a summary of all problems
    """
    if not filter:
        print(f"{8*'-'} we have a total of {len(problems)} problems")
    else:
        print(f"the problems whose name contains '{filter}' are:")
    for shortname, function in problems.items():
        problem = function()
        name = problem['name']
        if filter is not None and filter not in name and filter not in shortname:
            continue
        print_problem(problem, header=True)

def spot(needle):
    """
    returns the first problem instance whose name or shortname contains needle
    """
    for shortname, function in problems.items():
        problem = function()
        name = problem['name']
        if needle in name or needle in shortname:
            return problem


if __name__ == "__main__":
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument("-l", "--list", action="store_true")
    parser.add_argument("problem", nargs='?', default=None)
    args = parser.parse_args()
    if args.list:
        if args.problem:
            summary(args.problem)
        else:
            summary()
    elif args.problem:
        print_problem(spot(args.problem))
    else:
        summary()



import numpy as np
import timeit
from exact_cover import get_all_solutions
from xcover import covers_bool

matrix = np.genfromtxt("test_problem.txt", dtype = np.int32)

def solve_exact_cover():
    sols = get_all_solutions(matrix)
    print("exact_cover nsols:", len(sols))

def solve_xcover():
    sols = list(covers_bool(matrix))
    print("xcover nsols:", len(sols))

time_exact_cover = timeit.timeit(solve_exact_cover, number=5)
time_xcover = timeit.timeit(solve_xcover, number=5)

print(f"exact_cover solve in {time_exact_cover} s")
print(f"xcover solve in {time_xcover} s")
print(f"xcover faster by factor of {time_exact_cover/time_xcover}")




